from math import floor, sqrt
from pathlib import Path

import numpy as np
import pandas as pd
from scipy.stats import moment
from skimage.filters import sobel
from skimage.io import imsave
from skimage.measure import find_contours, shannon_entropy
from tifffile import imwrite

from model_server.extensions.chaeo.accessors import MonoPatchStack, Multichannel3dPatchStack
from model_server.extensions.chaeo.annotators import draw_box_on_patch, draw_contours_on_patch
from model_server.base.accessors import GenericImageDataAccessor, InMemoryDataAccessor
from model_server.base.process import pad, rescale, resample_to_8bit

def _make_rgb(zs):
    h, w, c, nz = zs.shape
    assert c <= 3
    outdata = np.zeros((h, w, 3, nz), dtype=zs.dtype)
    outdata[:, :, 0:c, :] = zs[:, :, :, :]
    return outdata

def _focus_metrics():
    return {
        'max_intensity': lambda x: np.max(x),
        'stdev': lambda x: np.std(x),
        'max_sobel': lambda x: np.max(sobel(x)),
        'rms_sobel': lambda x: sqrt(np.mean(sobel(x) ** 2)),
        'entropy': lambda x: shannon_entropy(x),
        'moment': lambda x: moment(x.flatten(), moment=2),
    }

def _safe_add(a, g, b):
    assert a.dtype == b.dtype
    assert a.shape == b.shape
    assert g >= 0.0

    return np.clip(
        a.astype('uint32') + g * b.astype('uint32'),
        0,
        np.iinfo(a.dtype).max
    ).astype(a.dtype)

def write_patch_to_file(where, fname, yxcz):
    ext = fname.split('.')[-1].upper()
    where.mkdir(parents=True, exist_ok=True)

    if ext == 'PNG':
        assert yxcz.dtype == 'uint8', f'Invalid data type {yxcz.dtype}'
        assert yxcz.shape[2] <= 3, f'Cannot export images with more than 3 channels as PNGs'
        assert yxcz.shape[3] == 1, f'Cannot export z-stacks as PNGs'
        if yxcz.shape[2] == 1:
            outdata = yxcz[:, :, 0, 0]
        elif yxcz.shape[2] == 2: # add a blank blue channel
            outdata = _make_rgb(yxcz)
        else: # preserve RGB order
            outdata = yxcz[:, :, :, 0]
        imsave(where / fname, outdata, check_contrast=False)
        return True

    elif ext in ['TIF', 'TIFF']:
        zcyx = np.moveaxis(yxcz, [3, 2, 0, 1], [0, 1, 2, 3])
        imwrite(where / fname, zcyx, imagej=True)
        return True

    else:
        raise Exception(f'Unsupported file extension: {ext}')



def get_patch_masks(roiset, pad_to: int = 256) -> MonoPatchStack:
    patches = []
    for roi in roiset.get_df().itertuples():
        patch = np.zeros((roi.ebb_h, roi.ebb_w, 1, 1), dtype='uint8')
        patch[roi.relative_slice][:, :, 0, 0] = roi.mask * 255

        if pad_to:
            patch = pad(patch, pad_to)

        patches.append(patch)
    return MonoPatchStack(patches)


def export_patch_masks(roiset, where: Path, pad_to: int = 256, prefix='mask', **kwargs) -> list:
    patches_acc = get_patch_masks(roiset, pad_to=pad_to)

    exported = []
    for i, roi in enumerate(roiset.get_df().itertuples()):  # assumes index of patches_acc is same as dataframe
        patch = patches_acc.iat_yxcz(i)
        ext = 'png'
        fname = f'{prefix}-la{roi.label:04d}-zi{roi.zi:04d}.{ext}'
        write_patch_to_file(where, fname, patch)
        exported.append(fname)
    return exported


def get_patches_from_zmask_meta(
        roiset,
        rescale_clip: float = 0.0,
        pad_to: int = 256,
        make_3d: bool = False,
        focus_metric: str = None,
        rgb_overlay_channels: list = None,
        rgb_overlay_weights: list = [1.0, 1.0, 1.0],
        white_channel: int = None,
        **kwargs
) -> pd.DataFrame:

    # arrange RGB channels if so specified, otherwise copy roiset.raw_acc data
    raw = roiset.acc_raw
    if isinstance(rgb_overlay_channels, list) and isinstance(rgb_overlay_weights, list):
        assert all([c < raw.chroma for c in rgb_overlay_channels])
        assert len(rgb_overlay_channels) == 3
        assert len(rgb_overlay_weights) == 3

        if white_channel:
            assert white_channel < raw.chroma
            stack = raw.data[:, :, [white_channel, white_channel, white_channel], :]
        else:
            stack = np.zeros([*raw.shape[0:1], 3, raw.shape[3]], dtype=raw.dtype)

        for ii, ci in enumerate(rgb_overlay_channels):
            if ci is None:
                continue
            assert isinstance(ci, int)
            assert ci < raw.chroma
            stack[:, :, ii, :] = _safe_add(
                stack[:, :, ii, :], # either black or grayscale channel
                rgb_overlay_weights[ii],
                stack[:, :, ci, :]
            )
    else:
        if white_channel:
            assert white_channel < raw.chroma
            stack = raw.data[:, :, [white_channel], :]
        else:
            stack = raw.data

    def _make_patch(roi):
        patch3d = stack[roi.slice]
        ph, pw, pc, pz = patch3d.shape
        subpatch = patch3d[roi.relative_slice]

        # make a 3d patch
        if make_3d:
            patch = patch3d

        # make a 2d patch, find optimal z-position determined by focus_metric function on each channel separately
        elif focus_metric is not None:
            foc = _focus_metrics()[focus_metric]

            patch = np.zeros([ph, pw, pc, 1], dtype=patch3d.dtype)

            for ci in range(0, pc):
                me = [foc(subpatch[:, :, ci, zi]) for zi in range(0, pz)]
                zif = np.argmax(me)
                patch[:, :, ci, 0] = patch3d[:, :, ci, zif]

        # make a 2d patch from middle of z-stack
        else:
            zim = floor(pz / 2)
            patch = patch3d[:, :, :, [zim]]

        assert len(patch.shape) == 4

        if rescale_clip is not None:
            patch = rescale(patch, rescale_clip)

        if kwargs.get('draw_bounding_box') is True:
            bci = kwargs.get('bounding_box_channel', 0)
            assert bci < 3
            if bci > 0:
                patch = _make_rgb(patch)
            for zi in range(0, patch.shape[3]):
                patch[:, :, bci, zi] = draw_box_on_patch(
                    patch[:, :, bci, zi],
                    ((roi.rel_x0, roi.rel_y0), (roi.rel_x1, roi.rel_y1)),
                    linewidth=kwargs.get('bounding_box_linewidth', 1)
                )

        if kwargs.get('draw_mask'):
            mci = kwargs.get('mask_channel', 0)
            mask = np.zeros(patch.shape[0:2], dtype=bool)
            mask[roi.relative_slice[0:2]] = roi.mask
            for zi in range(0, patch.shape[3]):
                patch[:, :, mci, zi] = np.invert(mask) * patch[:, :, mci, zi]

        if kwargs.get('draw_contour'):
            mci = kwargs.get('contour_channel', 0)
            mask = np.zeros(patch.shape[0:2], dtype=bool)
            mask[roi.relative_slice[0:2]] = roi.mask

            for zi in range(0, patch.shape[3]):
                patch[:, :, mci, zi] = draw_contours_on_patch(
                    patch[:, :, mci, zi],
                    find_contours(mask)
                )

        if pad_to:
            patch = pad(patch, pad_to)
        return patch

    dfe = roiset.get_df()
    dfe['patch'] = roiset.get_df().apply(lambda r: _make_patch(r), axis=1)
    return dfe


def export_patches_from_zstack(
        where: Path,
        roiset,
        prefix='patch',
        **kwargs
):
    make_3d = kwargs.get('make_3d', False)
    patches_df = get_patches_from_zmask_meta(roiset, **kwargs)

    pc = roiset.acc_raw.chroma
    patches = list(patches_df['patch'])
    if not make_3d and pc == 1:
        patches_acc = MonoPatchStack(patches)
    else:
        patches_acc = Multichannel3dPatchStack(patches)

    def _export_patch(roi):
        patch = patches_acc.iat_yxcz(i)
        ext = 'tif' if make_3d or patches_acc.chroma > 3 else 'png'
        fname = f'{prefix}-la{roi.label:04d}-zi{roi.zi:04d}.{ext}'

        if patch.dtype is np.dtype('uint16'):
            write_patch_to_file(where, fname, resample_to_8bit(patch))
        else:
            write_patch_to_file(where, fname, patch)

        exported.append({
            'df_index': roi.Index,
            'patch_filename': fname,
        })

    exported = []
    for i, roi in enumerate(patches_df.itertuples()):  # just used for label info
        _export_patch(roi)

    return exported


def export_multichannel_patches_from_zstack(
    where: Path,
    roiset,
    rgb_overlay_channels: list = None,
    rgb_overlay_weights: list = [1.0, 1.0, 1.0],
    white_channel: int = None,
    **kwargs
):
    """
    Export RGB patches where each patch is assignable to a channel of the input stack
    :param ch_rgb_overlay: tuple of integers (R, G, B) that assign a stack channel index to an RGB channel
    :param overlay_gain: optional, tuple of float (R, G, B) multipliers that can be used to balance relative brightness
    :param rgb_white_channel: int, index of stack channel that becomes grayscale signal in export patches
    """
    # def _safe_add(a, g, b):
    #     assert a.dtype == b.dtype
    #     assert a.shape == b.shape
    #     assert g >= 0.0
    #
    #     return np.clip(
    #         a.astype('uint32') + g * b.astype('uint32'),
    #         0,
    #         np.iinfo(a.dtype).max
    #     ).astype(a.dtype)
    #
    # idata = roiset.acc_raw.data
    # if rgb_white_channel:
    #     assert rgb_white_channel < roiset.acc_raw.chroma
    #     mdata = idata[:, :, [rgb_white_channel, rgb_white_channel, rgb_white_channel], :]
    # else:
    #     mdata = idata
    #
    # if rgb_overlay_channels:
    #     assert len(rgb_overlay_channels) == 3
    #     assert len(rgb_overlay_weights) == 3
    #     for ii, ci in enumerate(rgb_overlay_channels):
    #         if ci is None:
    #             continue
    #         assert isinstance(ci, int)
    #         assert ci < roiset.acc_raw.chroma
    #         mdata[:, :, ii, :] = _safe_add(
    #             mdata[:, :, ii, :],
    #             rgb_overlay_weights[ii],
    #             idata[:, :, ci, :]
    #         )
    #
    # # TODO: this is a bit of a workaround
    # mstack = InMemoryDataAccessor(mdata)
    # rgb_roiset = RoiSet(roiset.acc_obj_ids, mstack, roiset.params)
    # return export_patches_from_zstack(where, rgb_roiset, **kwargs)
    kwargs['white_channel'] = white_channel
    kwargs['rgb_overlay_channels'] = rgb_overlay_channels
    kwargs['rgb_overlay_weights'] = rgb_overlay_weights
    return export_patches_from_zstack(where, roiset, **kwargs)